#Python base libs
import sys

#Here it goes all the variables/constant that are shared between modules
CANVAS_SIZE = 0
DEBUG = False
DOWNSAMPLING = 1
REAL_CANVAS_SIZE = 0
ROOT = sys.path[0]


AVGTIME = []